
/*
C++ wrapper around Windows NT service control functions
Version 0.2
Copyright (C) 2000 Antti Markus (antti@pld.ttu.ee)     
This code is based on Windows Foundation Classes by Samuel R. Blackburn
(http://ourworld.compuserve.com/homepages/sam_blackburn/wfc.htm)

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
using std::string;
#include "svcmgr.h"

TServiceStatus::TServiceStatus()
{         
  Empty();
}

TServiceStatus::TServiceStatus(const ENUM_SERVICE_STATUS *pSource)
{               
  Copy(pSource);
}

TServiceStatus::~TServiceStatus()
{   
  Empty();
}

void TServiceStatus::Empty(void)
{                     
  lpServiceName=static_cast<LPSTR>(NULL);
  lpDisplayName=static_cast<LPSTR>(NULL);
  ServiceStatus.dwServiceType=0;
  ServiceStatus.dwCurrentState=0;
  ServiceStatus.dwControlsAccepted=0;
  ServiceStatus.dwWin32ExitCode=0;
  ServiceStatus.dwServiceSpecificExitCode=0;
  ServiceStatus.dwCheckPoint=0;
  ServiceStatus.dwWaitHint=0;
}

void TServiceStatus::Copy(const ENUM_SERVICE_STATUS *pSource)
{         
  if(pSource == NULL)
  {
    Empty();
    return;
  }        
  
  try
  {         
    lpServiceName=pSource->lpServiceName;
    lpDisplayName= pSource->lpDisplayName;
    ServiceStatus.dwServiceType=pSource->ServiceStatus.dwServiceType;
    ServiceStatus.dwCurrentState=pSource->ServiceStatus.dwCurrentState;
    ServiceStatus.dwControlsAccepted=pSource->ServiceStatus.dwControlsAccepted;
    ServiceStatus.dwWin32ExitCode=pSource->ServiceStatus.dwWin32ExitCode;
    ServiceStatus.dwServiceSpecificExitCode=pSource->ServiceStatus.dwServiceSpecificExitCode;
    ServiceStatus.dwCheckPoint=pSource->ServiceStatus.dwCheckPoint;
    ServiceStatus.dwWaitHint=pSource->ServiceStatus.dwWaitHint;    
  }
  catch(...)
  {
    Empty();
  }
} 

string TServiceStatus::GetState(void)
{                      
  switch(ServiceStatus.dwCurrentState)
  {
    case SERVICE_STOPPED:
      return "stopped";
      
    case SERVICE_RUNNING:
      return "running";
      
    default:
      return "unknown";    
  }
}

         
void TServiceControlManager::Init(void)
{    
  hManagerHandle=static_cast<SC_HANDLE>(NULL);
  dwErrorCode=0;
  pServiceStatusData=static_cast<ENUM_SERVICE_STATUS*>(NULL);
  dwBufferSize=0;
  dwCurrentEntryNumber=0;
  dwNumberOfEntries=0;
}         
         
TServiceControlManager::TServiceControlManager()
{                                                
  Init();
}

TServiceControlManager::~TServiceControlManager()
{                                                  
  Close();
  
  if(pServiceStatusData != NULL)
  {
    delete [] pServiceStatusData;
    pServiceStatusData=static_cast<ENUM_SERVICE_STATUS*>(NULL);
    dwBufferSize=0;
  }
} 

void TServiceControlManager::Close(void)
{                                
  if(hManagerHandle != NULL)
  {
    if(CloseServiceHandle(hManagerHandle) == FALSE)
    {
      dwErrorCode=GetLastError();
    }     
    hManagerHandle=static_cast<SC_HANDLE>(NULL);
  } //end if(pManagerHandle != NULL)
} 

BOOL TServiceControlManager::Open(DWORD dwAccess,
                                  LPCTSTR strDatabase,
                                  LPCTSTR strMachine)
{                                                   
  if(hManagerHandle != NULL)
  {
    Close();
  }         
  
  try
  {
    if(strMachine == NULL)
    {
      strMachineName="";
    }                  
    else
    {
      strMachineName=strMachine;
    }            
    
    hManagerHandle=OpenSCManager(strMachine, strDatabase, dwAccess);
    
    if(hManagerHandle == NULL)
    {
      dwErrorCode=GetLastError();
      return FALSE;
    }              
    else
    {
      return TRUE;
    }
  }
  catch(...)
  {                             
    dwErrorCode=ERROR_EXCEPTION_IN_SERVICE;
    return FALSE;
  }
}
  
BOOL TServiceControlManager::EnumerateStatus(DWORD dwState,
                                             DWORD dwType)
{                                                        
  if(hManagerHandle == NULL)
  {
    dwErrorCode=ERROR_INVALID_HANDLE;
    return FALSE;
  }              
  
  if(pServiceStatusData == NULL)
  {
    dwBufferSize=2*sizeof(ENUM_SERVICE_STATUS);
    
    try
    {
      //NB! Unsafe cast
      pServiceStatusData=reinterpret_cast<ENUM_SERVICE_STATUS*>(new BYTE[dwBufferSize]);
    }
    catch(...)
    {
      dwBufferSize=0;
      dwErrorCode=ERROR_NOT_ENOUGH_MEMORY;
      return FALSE;
    }  
    
  } //end if(pServiceStatusData == NULL
  
  DWORD dwBytesNeeded=0;
  DWORD dwServiceCount=0;
  DWORD dwResumeHandle=0;
  
  if(EnumServicesStatus(hManagerHandle,
                        dwType,
                        dwState,
                        pServiceStatusData,
                        dwBufferSize,
                        &dwBytesNeeded,
                        &dwServiceCount,
                        &dwResumeHandle) != FALSE)
  {
    dwCurrentEntryNumber=0;
    dwNumberOfEntries=dwServiceCount;

    return TRUE;
  }   
  
  dwErrorCode=GetLastError();
  
  if(dwErrorCode == ERROR_MORE_DATA)
  {                              
    delete [] pServiceStatusData;
    try       
    {                            
      //NB! Unsafe cast
      pServiceStatusData=reinterpret_cast<ENUM_SERVICE_STATUS*>(new BYTE[dwBytesNeeded+dwBufferSize]);
    }  
    catch(...)  
    {     
      dwBufferSize=0;
      dwErrorCode=ERROR_NOT_ENOUGH_MEMORY;
      return FALSE;
    }

    dwBufferSize+=dwBytesNeeded;
    
    dwBytesNeeded=0;
    dwServiceCount=0;
    dwResumeHandle=0;
    
    if(EnumServicesStatus(hManagerHandle,
                          dwType,
                          dwState,
                          pServiceStatusData,
                          dwBufferSize,
                          &dwBytesNeeded,
                          &dwServiceCount,
                          &dwResumeHandle) != FALSE)
    {
      dwCurrentEntryNumber=0;
      dwNumberOfEntries=dwServiceCount;
      return TRUE;
    }                                  
    else
    {
      dwErrorCode=GetLastError();
      dwCurrentEntryNumber=0;
      dwNumberOfEntries=0;
      return FALSE;
    }
  } //end if(dwErrorCode == ERROR_MORE_DATA
  
  return FALSE;
} //end EnumerateStatus
 
//NB! this method allocates memory which has to be freed later! 
ENUM_SERVICE_STATUS *TServiceControlManager::GetDependentServices(LPCTSTR pServiceName, DWORD &dwCount)
{                             
  if(pServiceName == NULL)
  {              
    dwErrorCode=ERROR_INVALID_PARAMETER;
    return NULL;
  }
                                 
  SC_HANDLE hService=static_cast<SC_HANDLE>(NULL);
  
  try
  {        
    hService=OpenService(hManagerHandle, pServiceName, SERVICE_ENUMERATE_DEPENDENTS);
    if(hService == NULL)
    {              
      dwErrorCode=GetLastError();
      return NULL;
    }      
    
    DWORD dwSvcCount=0;
    DWORD dwBytesNeeded=0;
    DWORD dwBufferSize=0;
    
    BOOL bReturnValue=EnumDependentServices(hService, 
                                            SERVICE_STATE_ALL, 
                                            static_cast<ENUM_SERVICE_STATUS*>(NULL),
                                            dwBufferSize,
                                            &dwBytesNeeded,
                                            &dwSvcCount);
    if(GetLastError() == ERROR_MORE_DATA)
    {   
      ENUM_SERVICE_STATUS *pBuf=static_cast<ENUM_SERVICE_STATUS*>(NULL); 
      try
      {                
        //NB! Unsafe cast
        pBuf=reinterpret_cast<ENUM_SERVICE_STATUS*>(new BYTE[dwBytesNeeded]);
      }
      catch(...)
      {              
        dwErrorCode=ERROR_NOT_ENOUGH_MEMORY;
        return NULL;
      }                                        
      ZeroMemory(pBuf, dwBytesNeeded);
      dwBufferSize=dwBytesNeeded;
      dwBytesNeeded=0;
      
      bReturnValue=EnumDependentServices(hService,
                                         SERVICE_STATE_ALL,
                                         pBuf,
                                         dwBufferSize,
                                         &dwBytesNeeded,
                                         &dwSvcCount);  
                                         
      if(bReturnValue == FALSE)
      {              
        dwErrorCode=GetLastError();
        delete [] pBuf;
        return NULL;
      }           
      
      dwCount=dwSvcCount;                   
      return pBuf;
    }
    else
    {              
      dwErrorCode=GetLastError();
      return NULL;
    }                                            
  }
  catch(...)
  { 
    dwErrorCode=ERROR_EXCEPTION_IN_SERVICE;
    return NULL;
  }
} //end GetDependentServices

BOOL TServiceControlManager::GetNext(TServiceStatus &Status)
{              
  if(dwCurrentEntryNumber < dwNumberOfEntries)
  {
    Status.Copy(&pServiceStatusData[dwCurrentEntryNumber]);
    dwCurrentEntryNumber++;
    return TRUE;
  }

  return FALSE;
}

BOOL TServiceControlManager::Start(LPCTSTR pServiceName, DWORD dwArgumentCount, LPCTSTR *pArguments)
{                            
  if(pServiceName == NULL)
  {              
    dwErrorCode=ERROR_INVALID_PARAMETER;
    return FALSE;
  }                                     
  
  try
  {              
    if(hManagerHandle == NULL)
    {
      dwErrorCode=ERROR_INVALID_HANDLE;
      return FALSE;
    }
    
    SC_HANDLE hService=OpenService(hManagerHandle, pServiceName, SERVICE_START);
    if(hService == static_cast<SC_HANDLE>(NULL))
    {
      dwErrorCode=GetLastError();
      return FALSE;
    }              
    
    BOOL bResult=StartService(hService, dwArgumentCount, pArguments);
    if(bResult == FALSE)
    {
      dwErrorCode=GetLastError();
      CloseServiceHandle(hService);
      return FALSE;
    }              
    
    CloseServiceHandle(hService);
    return TRUE;
  }
  catch(...)
  {
    dwErrorCode=ERROR_EXCEPTION_IN_SERVICE;
    return FALSE;
  }
} //end Start

BOOL TServiceControlManager::Stop(LPCTSTR pServiceName)
{                            
  if(pServiceName == NULL)
  {             
    dwErrorCode=ERROR_INVALID_PARAMETER;
    return FALSE;
  }                                     
  
  try
  {              
    if(hManagerHandle == NULL)
    {
      dwErrorCode=ERROR_INVALID_HANDLE;
      return FALSE;
    }              
    
    SC_HANDLE hService=OpenService(hManagerHandle, pServiceName, SERVICE_STOP | SERVICE_QUERY_STATUS);
    if(hService == static_cast<SC_HANDLE>(NULL))
    {
      dwErrorCode=GetLastError();
      return FALSE;
    }              
    
    SERVICE_STATUS StatusInfo;
    
    BOOL bResult=ControlService(hService, SERVICE_CONTROL_STOP, &StatusInfo);
    if(bResult == FALSE)
    {
      dwErrorCode=GetLastError();
      CloseServiceHandle(hService);
      return FALSE;
    }              
    
    DWORD dwPreviousCheckpoint=0;
    DWORD dwTryCount=0;
    
    while(StatusInfo.dwCurrentState != SERVICE_STOPPED)
    {                                          
      dwPreviousCheckpoint=StatusInfo.dwCheckPoint;
      
      Sleep(StatusInfo.dwWaitHint);
      
      if(QueryServiceStatus(hService, &StatusInfo) == FALSE)
      {                                          
        CloseServiceHandle(hService);
        dwErrorCode=GetLastError();
        return FALSE;
      }                            
      
      if(dwPreviousCheckpoint >= StatusInfo.dwCheckPoint)
      {
        dwTryCount++;
        
        if(dwTryCount >= 1000)
        {
          CloseServiceHandle(hService);
          return FALSE;
        }
      }                
      else
      {
        dwTryCount=0;
      }
    } //end while(StatusInfo.dwCurrentState ...
    
    CloseServiceHandle(hService);
    return TRUE;
  }
  catch(...)
  {
    dwErrorCode=ERROR_EXCEPTION_IN_SERVICE;
    return FALSE;
  }
}
                                  